<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace AxiosAssystModule\Command;

use App\itnovum\openITCOCKPIT\Monitoring\Naemon\ExternalCommands;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use App\Model\Table\SystemsettingsTable;
use AxiosAssystModule\Model\Table\AxiosObjectsTable;
use Cake\Console\Arguments;
use Cake\Console\Command;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Core\Configure;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\Log\Log;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use itnovum\openITCOCKPIT\Core\Views\Host;
use itnovum\openITCOCKPIT\Core\Views\HoststatusIcon;
use itnovum\openITCOCKPIT\Core\Views\Service;
use itnovum\openITCOCKPIT\Core\Views\ServicestatusIcon;


/**
 * AxiosAssystNotification command.
 *
 * Beschreibung:
 * Das Script erzeugt Tickets über einen SOAP Service.
 * Konnte ein Ticket erfolgreich erzeugt werden, ist die Response von Axios Assyst die Ticket-Referenz-ID.
 * Ist dies der Fall wird ein Acknowlegdement mit der Ticket-Referenz-ID in openITCOCKPIT ausgelöst.
 * Wird ein Ticket geschlossen ist eine URL durch Axios Assyst aufzurufen, die die Ticket-Referenz-ID aus dem
 * openITCOCKPIT entfernt.
 *
 * @link https://servicedesk.fraunhofer.de/assystweb/event/DisplayEvent.do?dispatch=getEvent&eventId
 */
class AxiosAssystNotificationCommand extends Command {


    /**
     * Axios Assyst SOAP login data
     * Gets defined in AxiosAssystModule/config/webservice.php
     *
     * @var array
     */
    protected $config = [];

    /**
     * @var ConsoleIo
     */
    protected $io;

    private $systemAddress = '';

    private $type = 'host';

    private $hostUuid = '';

    private $serviceUuid = '';

    /**
     * 0 = Up, 1 = Down, 2 = Unreachable
     *
     * @var null|int
     */
    private $hoststate = null;

    /**
     * 0 = Ok, 1 = Warning, 2 = Critical, 3 = Unknown
     *
     * @var null|int
     */
    private $servicestate = null;

    private $hostoutput = '';

    private $serviceoutput = '';

    /**
     * PROBLEM", "RECOVERY", "ACKNOWLEDGEMENT", "FLAPPINGSTART", "FLAPPINGSTOP",
     * "FLAPPINGDISABLED", "DOWNTIMESTART", "DOWNTIMEEND", or "DOWNTIMECANCELLED"
     *
     * @var string
     */
    private $notificationtype = '';


    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/3.0/en/console-and-shells/commands.html#defining-arguments-and-options
     *
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        $parser->addOptions([
            'type'             => ['help' => __d('oitc_console', 'Host or Service')],
            'notificationtype' => ['help' => __d('oitc_console', 'PROBLEM')],
            'hostuuid'         => ['help' => __d('oitc_console', 'The Host UUID')],
            'hostoutput'       => ['help' => __d('oitc_console', 'The Host output')],
            'hoststate'        => ['help' => __d('oitc_console', 'State ID of the current host state')],
            'serviceuuid'      => ['help' => __d('oitc_console', 'The Service UUID')],
            'serviceoutput'    => ['help' => __d('oitc_console', 'The Service output')],
            'servicestate'     => ['help' => __d('oitc_console', 'State ID of the current service state')]
        ]);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     * @throws \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        $this->io = $io;
        Configure::load('AxiosAssystModule.webservice');
        $this->config = Configure::read('axios');

        $this->validateOptions($args);

        /** @var SystemsettingsTable $SystemsettingsTable */
        $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
        $_systemsettings = $SystemsettingsTable->findAsArray();

        $this->systemAddress = $_systemsettings['SYSTEM']['SYSTEM.ADDRESS'];

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        try {
            $host = $HostsTable->getHostByUuid($this->hostUuid, false);
        } catch (RecordNotFoundException $e) {
            throw new \RuntimeException(sprintf('Host with uuid "%s" could not be found!', $this->hostUuid));
        }

        $Host = new Host($host);

        if ($this->checkWSDLURL()) {
            if ($this->type === 'service') {
                /** @var ServicesTable $ServicesTable */
                $ServicesTable = TableRegistry::getTableLocator()->get('Services');

                try {
                    $service = $ServicesTable->getServiceByUuid($this->serviceUuid, false);
                } catch (RecordNotFoundException $e) {
                    throw new \RuntimeException(sprintf('Service with uuid "%s" could not be found!', $this->serviceUuid));
                }

                $Service = new Service($service);
                $this->sendServiceNotification($Host, $Service);
                exit(0);
            }


            $this->sendHostNotification($Host);
            exit(0);
        }
        exit(1);
    }

    /**
     * @return bool
     */
    private function checkWSDLURL() {
        try {
            $client = new \SoapClient($this->config['webserviceURL']);
            return true;
        } catch (\SoapFault $fault) {
            $this->io->error(sprintf(
                'SOAP Fault [%s]: %s)',
                $fault->faultcode,
                $fault->faultstring
            ));
        }
        return false;
    }

    /**
     * returns the real reference id since axios assyst ids consists out of the reference id + 10mio
     *
     * @param int|string $refId
     *
     * @return int
     */
    private function getRealRefId($refId) {
        $refId = (int)$refId;
        return 10000000 + $refId;
    }


    /**
     * returns the axios assyst data in xml representation
     *
     * @param $array - the array which you want to get as XML
     *
     * @return string
     */
    private function assystDataAsXML($array) {
        $data = array_flip($array);
        $xml = new \SimpleXMLElement('<?xml version="1.0" encoding="UTF-8"?><assyst_data/>');
        array_walk_recursive($data, [$xml, 'addChild']);

        return $this->removeBOM($xml->asXML());
    }

    /**
     *
     * Queries the soap interface on the axios assyst system
     *
     * @param string $title
     * @param Host $Host
     * @param Service|null $Service
     */
    private function soapCall(string $title, Host $Host, ?Service $Service = null) {
        try {

            $client = new \SoapClient($this->config['webserviceURL'], [
                'login'    => $this->config['webserviceUser'],
                'password' => $this->config['webservicePass'],
                'trace'    => true,
            ]);

            //check if ref id exists

            $elementId = $Host->getId();
            if ($this->type === 'service') {
                $elementId = $Service->getId();
            }
            $data = $this->getReferenceData($elementId);

            $detailedDescription = 'https://' . $this->systemAddress . '/#!/hosts/browser/' . $elementId;
            if ($this->type === 'service') {
                $detailedDescription = 'https://' . $this->systemAddress . '/#!/services/browser/' . $elementId;
            }

            //may can be done with the getEleemntInfo fn

            $refId = null;
            if (!empty($data)) {
                $refId = $data['axios_id'];
            }

            if (!empty($refId) && is_numeric($refId)) {
                //reference ID exists

                //Merge assyst data with default assyst config
                $assystData = Hash::merge(
                    $this->config['assyst_data'],
                    [
                        'assystEventRef' => $refId,
                        'SystemEvent'    => 'GETSTATUS'
                    ]
                );

                $assystData = $this->assystDataAsXML($assystData);
                Log::info('AxiosAssyst Notification: ' . $assystData);

                if (in_array($client->assystEvent($assystData), ['close', 'geschlossen', 'closed', 'anstehend-schliessen'], true)) {
                    //ticket is closed so create a new one

                    $assystData = Hash::merge(
                        $this->config['assyst_data'],
                        [
                            'ShortDescription'    => $title,
                            'DetailedDescription' => $detailedDescription,
                        ]
                    );

                    $refId = $client->assystEvent($this->assystDataAsXML($assystData));
                    if (!empty($refId)) {
                        Log::info('AxiosAssyst Notification:  New support ticket has been successfully created');
                        $this->saveReferenceId($refId, $elementId);
                    } else {
                        Log::error('AxiosAssyst Notification:  Could not create new support ticket!');
                    }
                } else {
                    //state "offen"
                    //ticket is still open so refresh Ticketdata

                    $assystData = Hash::merge(
                        $this->config['assyst_data'],
                        [
                            'assystEventRef'      => $refId,
                            'DetailedDescription' => $detailedDescription,
                            'SystemEvent'    => 'MEHR-INFO'
                        ]
                    );

                    $assystData = $this->assystDataAsXML($assystData);
                    Log::info('AxiosAssyst Notification Assyst data: ' . $assystData);

                    $ticketData = $client->assystEvent($assystData);
                    Log::info('AxiosAssyst Notification Ticket data: ' . json_encode($ticketData));
                    if (empty($ticketData)) {
                        Log::info('Support ticket has been successfully updated:' . $refId);
                    }
                }

            } else if ($refId === null && $this->notificationtype === 'PROBLEM') {
                $assystData = Hash::merge(
                    $this->config['assyst_data'],
                    [
                        'ShortDescription'    => $title,
                        'DetailedDescription' => $detailedDescription
                    ]
                );

                Log::info('AxiosAssyst Notification: ' . $this->assystDataAsXML($assystData));

                //reference ID doesnt exist so create a new ticket

                $refId = $client->assystEvent($this->assystDataAsXML($assystData));
                Log::info('AxiosAssyst Notification: New Ref id: ' . $refId);


                if (!empty($refId) && is_numeric($refId)) {
                    Log::info('AxiosAssyst Notification: New support ticket has been successfully created');
                } else {
                    Log::error('AxiosAssyst Notification: Could not create new support ticket!');
                }
            }

            if (!empty($refId)) {
                $this->sendMonitoringACK($Host, $Service, $refId);
            }
        } catch (\Exception $e) {
            Log::error('AxiosAssyst Notification: ' . $e->getMessage());
        }
    }

    /**
     * @param Host $Host
     * @param Service|null $Service
     * @param $refId
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     */
    private function sendMonitoringACK(Host $Host, ?Service $Service, $refId) {
        $ExternalCommands = new ExternalCommands();

        $data = [
            'author'  => 'Axios Assyst',
            'comment' => 'Ticket_' . $refId . ';' . $this->getRealRefId($refId),
        ];

        if ($this->type === 'host') {
            $data = [
                'hostUuid' => $Host->getUuid(),
                'author'   => $data['author'],
                'comment'  => $data['comment'],
                'sticky'   => 2, // 2 = sticky with Nagios Core, 1 = Sticky with Naemon Core, 2 works for both :)
                'type'     => 'hostOnly',
            ];
            $ExternalCommands->setHostAck($data);
        } else {
            $data = [
                'hostUuid'    => $Host->getUuid(),
                'serviceUuid' => $Service->getUuid(),
                'author'      => $data['author'],
                'comment'     => $data['comment'],
                'sticky'      => 2 // 2 = sticky with Nagios Core, 1 = Sticky with Naemon Core, 2 works for both :)
            ];
            $ExternalCommands->setServiceAck($data);
        }
    }

    /**
     * Saves the reference ID of Axios Assyst ticket in the axios_objects table
     *
     * @param int $id
     * @param int $elementId
     */
    private function saveReferenceId($id, $elementId) {
        $type = lcfirst($this->type);

        /** @var AxiosObjectsTable $AxiosObjectsTable */
        $AxiosObjectsTable = TableRegistry::getTableLocator()->get('AxiosAssystModule.AxiosObject');

        // Delete old Reference Id
        $oldObjects = $AxiosObjectsTable
            ->find()
            ->where([
                'element_id' => $elementId,
                'type'       => $type
            ])
            ->all();

        foreach ($oldObjects as $oldObject) {
            Log::info('AxiosAssyst Notification: Remove old axios object reference: ' . $oldObject->id);
            $AxiosObjectsTable->delete($oldObject);
        }

        //Save new reference
        $entity = $AxiosObjectsTable->newEntity([
            'axios_id'   => $id,
            'element_id' => $elementId,
            'type'       => $type,
        ]);

        $AxiosObjectsTable->save($entity);

        Log::info('AxiosAssyst Notification: Save new reference: ' . $entity->id);
    }

    /**
     * @param int $elementId
     * @return array
     */
    private function getReferenceData($elementId) {
        $type = lcfirst($this->type);

        /** @var AxiosObjectsTable $AxiosObjectsTable */
        $AxiosObjectsTable = TableRegistry::getTableLocator()->get('AxiosAssystModule.AxiosObject');

        try {
            $result = $AxiosObjectsTable
                ->find()
                ->where([
                    'element_id' => $elementId,
                    'type'       => $type
                ])
                ->disableHydration()
                ->firstOrFail();

            return $result;
        } catch (RecordNotFoundException $e) {
            Log::error('AxiosAssyst Notification: No reference data found in database');
            return [];
        }
    }


    /**
     * Removed Byte order Mark from the XML
     *
     * @param string $str
     * @return string|string[]|null
     */
    private function removeBOM($str) {
        $val = preg_replace('/^(\x00\x00\xFE\xFF|\xFF\xFE\x00\x00|\xFE\xFF|\xFF\xFE|\xEF\xBB\xBF)/', "", $str);
        if (empty($val)) {
            return $str;
        }

        return $val;
    }

    private function isAcknowledgement() {
        return $this->notificationtype === 'ACKNOWLEDGEMENT';
    }

    private function isFlappingStart() {
        return $this->notificationtype === 'FLAPPINGSTART';
    }

    private function isFlappingStop() {
        return $this->notificationtype === 'FLAPPINGSTOP';
    }

    private function isFlappingDisabled() {
        return $this->notificationtype === 'FLAPPINGDISABLED';
    }

    private function isDowntimeStart() {
        return $this->notificationtype === 'DOWNTIMESTART';
    }

    private function isDowntimeEnd() {
        return $this->notificationtype === 'DOWNTIMEEND';
    }

    private function isDowntimeCancelled() {
        return $this->notificationtype === 'DOWNTIMECANCELLED';
    }

    private function sendHostNotification(Host $Host) {
        $HoststatusIcon = new HoststatusIcon($this->hoststate);

        if ($this->isAcknowledgement()) {
            $title = sprintf(
                'Acknowledgement for %s (%s)',
                $Host->getHostname(),
                $HoststatusIcon->getHumanState()
            );

            return $this->soapCall($title, $Host);
        }

        if ($this->isDowntimeStart()) {
            $title = sprintf(
                'Downtime start for %s',
                $Host->getHostname()
            );

            $message = 'Host has entered a period of scheduled downtime';
            return $this->soapCall($title, $Host);
        }

        if ($this->isDowntimeEnd()) {
            $title = sprintf(
                'Downtime end for %s',
                $Host->getHostname()
            );

            $message = 'Host has exited from a period of scheduled downtime';
            return $this->soapCall($title, $Host);
        }

        if ($this->isDowntimeCancelled()) {
            $title = sprintf(
                'Downtime cancelled for %s',
                $Host->getHostname()
            );

            $message = 'Scheduled downtime for host has been cancelled';
            return $this->soapCall($title, $Host);
        }

        if ($this->isFlappingStart()) {
            $title = sprintf(
                'Flapping started on %s',
                $Host->getHostname()
            );

            $message = 'Host appears to have started flapping';
            return $this->soapCall($title, $Host);
        }

        if ($this->isFlappingStop()) {
            $title = sprintf(
                'Flapping stopped on %s',
                $Host->getHostname()
            );

            $message = 'Host appears to have stopped flapping';
            return $this->soapCall($title, $Host);
        }

        if ($this->isFlappingDisabled()) {
            $title = sprintf(
                'Disabled flap detection for %s',
                $Host->getHostname()
            );

            $message = 'Flap detection has been disabled';
            return $this->soapCall($title, $Host);
        }

        //Default notification
        $title = sprintf(
            '%s: %s is %s!',
            $this->notificationtype,
            $Host->getHostname(),
            $HoststatusIcon->getHumanState()
        );

        return $this->soapCall($title, $Host);
    }

    private function sendServiceNotification(Host $Host, Service $Service) {
        $ServicestatusIcon = new ServicestatusIcon($this->servicestate);

        if ($this->isAcknowledgement()) {
            $title = sprintf(
                'Acknowledgement for %s/%s (%s)',
                $Host->getHostname(),
                $Service->getServicename(),
                $ServicestatusIcon->getHumanState()
            );

            return $this->soapCall($title, $Host, $Service);
        }

        if ($this->isDowntimeStart()) {
            $title = sprintf(
                'Downtime start for %s/%s',
                $Host->getHostname(),
                $Service->getServicename()
            );

            $message = 'Service has entered a period of scheduled downtime';
            return $this->soapCall($title, $Host, $Service);
        }

        if ($this->isDowntimeEnd()) {
            $title = sprintf(
                'Downtime end for %s/%s',
                $Host->getHostname(),
                $Service->getServicename()
            );

            $message = 'Service has exited from a period of scheduled downtime';
            return $this->soapCall($title, $Host, $Service);
        }

        if ($this->isDowntimeCancelled()) {
            $title = sprintf(
                'Downtime cancelled for %s/%s',
                $Host->getHostname(),
                $Service->getServicename()
            );

            $message = 'Scheduled downtime for service has been cancelled';
            return $this->soapCall($title, $Host, $Service);
        }

        if ($this->isFlappingStart()) {
            $title = sprintf(
                'Flapping started on %s/%s',
                $Host->getHostname(),
                $Service->getServicename()
            );

            $message = 'Service appears to have started flapping';
            return $this->soapCall($title, $Host, $Service);
        }

        if ($this->isFlappingStop()) {
            $title = sprintf(
                'Flapping stopped on %s/%s',
                $Host->getHostname(),
                $Service->getServicename()
            );

            $message = 'Service appears to have stopped flapping';
            return $this->soapCall($title, $Host, $Service);
        }

        if ($this->isFlappingDisabled()) {
            $title = sprintf(
                'Disabled flap detection for %s/%s',
                $Host->getHostname(),
                $Service->getServicename()
            );

            $message = 'Flap detection has been disabled';
            return $this->soapCall($title, $Host, $Service);
        }

        //Default notification
        $title = sprintf(
            '%s: %s on %s is %s!',
            $this->notificationtype,
            $Service->getServicename(),
            $Host->getHostname(),
            $ServicestatusIcon->getHumanState()
        );

        return $this->soapCall($title, $Host, $Service);
    }

    /**
     * @param Arguments $args
     * @throws \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions
     */
    private function validateOptions(Arguments $args) {
        if ($args->getOption('type') === '') {
            throw new \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions(
                'Option --type is missing'
            );
        }

        $this->type = strtolower($args->getOption('type'));

        if ($args->getOption('notificationtype') === '') {
            throw new \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions(
                'Option --notificationtype is missing'
            );
        }
        $this->notificationtype = $args->getOption('notificationtype');

        if ($args->getOption('hostuuid') === '') {
            throw new \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions(
                'Option --hostuuid is missing'
            );
        }
        $this->hostUuid = $args->getOption('hostuuid');

        if ($this->type === 'host') {
            $this->hoststate = 1;
            if ($args->getOption('hostoutput') === '') {
                throw new \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions(
                    'Option --hostoutput is missing'
                );
            }
            $this->hostoutput = $args->getOption('hostoutput');
        }


        if ($this->type === 'service') {
            $this->servicestate = 3;
            if ($args->getOption('serviceuuid') === '') {
                throw new \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions(
                    'Option --serviceuuid is missing'
                );
            }
            $this->serviceUuid = $args->getOption('serviceuuid');

            if ($args->getOption('serviceoutput') === '') {
                throw new \itnovum\openITCOCKPIT\ApiShell\Exceptions\MissingParameterExceptions(
                    'Option --serviceoutput is missing'
                );
            }
            $this->serviceoutput = $args->getOption('serviceoutput');
        }

        if ($args->getOption('hoststate') !== '') {
            //Not all notifications have a state like ack or downtime messages.
            $this->hoststate = (int)$args->getOption('hoststate');
        }

        if ($args->getOption('servicestate') !== '') {
            //Not all notifications have a state like ack or downtime messages.
            $this->servicestate = (int)$args->getOption('servicestate');
        }
    }
}
